// Copyright 2021, Paul Mosier
//
// This file is part of Scriptura, a ncurses-based Bible study
// software for the libsword backend.
//
// Scriptura is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, under version 2 of the License.
//
// Scriptura is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Scriptura.  If not, see <https://www.gnu.org/licenses/>.

// system
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ncurses.h>
#include <unistd.h>
#include <locale.h>

// third-party
#include <swconfig.h>

// mine
#include "free.h"
#include "pane.h"
#include "scabbard.h"


/* --[ ASSISTS / GLOBALS ]-- */

int NUMPANES;
int halfcols, halflines;
int floatx, floaty, floatheight, floatwidth;
int startx1, starty1, startx2, starty2;
int panelength, panewidth;

sword::SWConfig config;

//! Refresh every portion of a list of panes we are given.
void wipeit(pane* panes) {
	for (int i = 0; i < NUMPANES; i++) panes[i].redraw(true);
}

//! Display a form field.
starray showForm(const char* title, starray inputs, const char* secondinst,
								 int floaty, int floatx, int floatheight, int floatwidth) {
	// pull up the floating form
	pane form = {floaty, floatx, floatheight, floatwidth, title};

	starray ret = form.loadForm(inputs, secondinst);
	// sanitize input? -- if (ret.length != 0)
	return ret;
}

//! Save the config array -- we abstract this out in case this changes.
void saveit(char* configfile) {
	/* This is clumsy but we have to use this particular constructor for
	 * config() so we can save to disk.  The global config object uses a
	 * different constructor. */
	sword::SWConfig newconfig(configfile);
	newconfig.augment(config);
	newconfig.save();
}

//! Set various standard screen landmarks.
void landmarks() {
	// place background text
	mvprintw(LINES - 1 , COLS - 20, "%s", "? - help   q - quit");
	refresh();

	// split up the screen
	halfcols = COLS / 2;
	halflines = LINES / 2;

	// floating window measurements
	floatx = 3;
	floaty = 3;
	floatheight = LINES - 6;
	floatwidth = COLS - 6;

	// define our main windows and their borders
	if (! strcmp(config["layout"]["panels"], "v")) {
		// vertical layout
		startx1 = starty1 = startx2 = 0;
		starty2 = halflines;
		panelength = LINES - halflines - 1;
		panewidth = COLS;
	} else {
		// horizontal layout
		startx1 = starty1 = starty2 = 0;
		startx2 = halfcols;
		panelength = LINES - 1;
		panewidth = COLS - halfcols;
	}
}

//! Resize the standard screen.
void doresize(pane* p) {
	// terminal window was resized
	clear();

	// reset standard screen landmarks
	landmarks();

	// reset and refresh main panes
	p[0].resize(starty1, startx1, panelength, panewidth);
	p[1].resize(starty2, startx2, panelength, panewidth);
	wipeit(p);
}


/* --[ MAIN ]-- */
int main(int argc, char** argv) {
	// if help argument supplied, give text & exit
	if ((argc > 1) &&
			((! strcmp(argv[1], "-h")) || (! strcmp(argv[1], "--help")))) {
		printf("%s\n", CLITEXT);
		exit(0);
	}

	// read in our configuration file, or whatever the user supplies
	char* homedir = getenv("HOME");
	if (homedir == NULL) {
		printf("No home directory, pilgrim.\n");
		perror("(Could not find $HOME.)\n");
		return -1;
	}
	char* configfile;
	asprintf(&configfile, "%s%s", homedir, "/.config/scriptura.ini");

	if ((argc > 2) && (! strcmp(argv[1], "-c"))) {
		free(configfile);
		configfile = argv[2];

		if (access(configfile, R_OK|W_OK) == -1) {
			perror("Error loading supplied configuration file");
			return -1;
		}

	} else {
		// configfile remains default
		if (access(configfile, R_OK|W_OK) == -1) {
			printf("Creating new configuration file...");
			FILE *cnf = fopen(configfile, "w");
			if (cnf == NULL) {
				printf("\nCould not create new config file $HOME/.config/scriptura.ini\n");
				printf("Please ensure this directory exists and is writable.\n");
				return -1;
			}

			// set default module and other fields to avoid crashing
			fprintf(cnf, "[defaults]\n");
			fprintf(cnf, "scope=Gen 1:1 - Rev 22:21\n");
			fprintf(cnf, "searchkey=2 Tim 1:7\n");
			fprintf(cnf, "module=KJV\n");

			fclose(cnf);
			printf("done.\n");
		}
	}
	sword::SWConfig tempconfig(configfile);
	config.augment(tempconfig);

	// ensure some needed config settings are in place
	if (! strcmp(config["defaults"]["scope"], ""))
		config["defaults"]["scope"] = "Gen 1:1 - Rev 22:21";
	if (! strcmp(config["defaults"]["searchkey"], ""))
		config["defaults"]["searchkey"] = "2 Tim 1:;7";
	if (! strcmp(config["defaults"]["module"], ""))
		config["defaults"]["module"] = "KJV";
	saveit(configfile);

	// NOTE -- anything after this point must use our defined exit wrapup()

	/* start ncurses, disable line buffering hide cursor, and allow for fancy
	 * keys & so on */
	initscr();
	cbreak();
	curs_set(0);
	noecho();
	start_color();
	keypad(stdscr, true);
	setlocale(LC_ALL, "");

	// get color settings
	if (! strcmp(config["markup"]["lettercolor"], "green")) {
		init_pair(1, COLOR_GREEN, COLOR_BLACK);
	} else if (! strcmp(config["markup"]["lettercolor"], "yellow")) {
		init_pair(1, COLOR_YELLOW, COLOR_BLACK);
	} else if (! strcmp(config["markup"]["lettercolor"], "magenta")) {
		init_pair(1, COLOR_MAGENTA, COLOR_BLACK);
	} else if (! strcmp(config["markup"]["lettercolor"], "blue")) {
		init_pair(1, COLOR_BLUE, COLOR_BLACK);
	} else if (! strcmp(config["markup"]["lettercolor"], "cyan")) {
		init_pair(1, COLOR_CYAN, COLOR_BLACK);
	} else {
		init_pair(1, COLOR_RED, COLOR_BLACK);
	}

	// two main panes
	NUMPANES = 2;
	landmarks();
	pane* p = (pane*) malloc(sizeof(pane) * NUMPANES);
	const char* title = "Window";
	if (! p) wrapup(1, "Error allocating memory for panes.\n");
	p[0] = { starty1, startx1, panelength, panewidth, title };
	p[1] = { starty2, startx2, panelength, panewidth, title };

	pane* focustab;
	focustab = &(p[0]);
	focustab->toggleFocus();

	/* load up a scabbard & set blank starting data - scabbard args are
	 * sanitized in the constructor */
	scabbard scab = {};
	char blank[] = "Open a text to use this window.\0";
	for (int i = 0; i < NUMPANES; i++) {
		if (config["defaults"]["module"] && config["defaults"]["searchkey"]) {
			// this isn't technically the title, but it's what we have now
			p[i].setTitle(config["defaults"]["module"]);
			p[i].loadText(scab.getSpan(p[i].getModkey()));
		} else {
			p[i].loadText(blank);
		}
	}
	wipeit(p);

	// loop for input
	int ch;
	while ((ch = getch()) != 'q') {
		switch (ch) {

		case '\t': {
			// switch pane focus
			focustab->toggleFocus();
			focustab = (focustab == &(p[0]) ? &(p[1]) : &(p[0]));
			focustab->toggleFocus();
			break; }

		case KEY_NPAGE:
			focustab->nextPage();
			break;

		case KEY_DOWN:
			focustab->scrollDown();
			break;

		case KEY_UP:
			focustab->scrollUp();
			break;

		case KEY_PPAGE:
			focustab->prevPage();
			break;

		case KEY_RESIZE:
			doresize(p);
			break;

		case 'f': {
			// toggle footnotes
			int footnotes = ! parseConf(config["markup"]["footnotes"]);
			config["markup"]["footnotes"] = footnotes + '0';
			wipeit(p);
			saveit(configfile);
			break; }

		case 'g': {
			// go to a particular passage
			starray goinst;
			goinst = stinit(goinst);
			goinst = stappend(goinst, "Go to: ");

			const char* gotitle = "Go to passage";
			const char* secondinst = "Enter passage to retrieve; "
				"abbreviations & disjoint passage selections are okay.";

			starray ret = showForm(gotitle, goinst, secondinst, floaty,
														 floatx, floatheight, floatwidth);
			if (ret.length == -1) {
				doresize(p);
				break;
			}

			if ((ret.length != 0) && (strlen((ret.strings)[0]) > 0)) {
				// user entered something that's not a blank string
				focustab->setKey((ret.strings)[0]);
				focustab->loadText(scab.getSpan(focustab->getModkey()));
			}
			wipeit(p);
			break; }

		case 'n': {
			// toggle non-textual words
			int textual = ! parseConf(config["markup"]["nontextual"]);
			config["markup"]["nontextual"] = textual + '0';
			wipeit(p);
			saveit(configfile);
			break; }

		case 'o': {
			// load a module into the current pane
			const char* typetitle = "Choose a module type";
			pane menupane = {floaty, floatx, floatheight, floatwidth,
				typetitle };
			starray mc = scab.getModClassifications();
			int modclass = menupane.loadMenu(mc);
			if (modclass == -1) {
				wipeit(p);
				break;
			} else if (modclass == -2) {
				doresize(p);
				break;
			}

			const char* modtitle = "Choose a module";
			menupane = {floaty, floatx, floatheight, floatwidth, modtitle};
			starray mds = scab.getModDescriptions(modclass);
			int mod = menupane.loadMenu(mds);
			if (mod == -1) {
				wipeit(p);
				break;
			} else if (mod == -2) {
				doresize(p);
				break;
			}

			focustab->setModule(scab.getModDescription(modclass, mod),
													scab.getModName(modclass, mod),
													scab.getKeyType(modclass));
			focustab->loadText(scab.getSpan(focustab->getModkey()));
			wipeit(p);
			break; }

		case 'r': {
			// toggle redletter
			int redletter = ! parseConf(config["markup"]["redletters"]);
			config["markup"]["redletters"] = redletter + '0';
			wipeit(p);
			saveit(configfile);
			break; }

		case 's': {
			// search
			starray inst;
			inst = stinit(inst);

			/* NOTE -- these happen to be in the same order as the
			 * integer key used by Sword to determine the search type,
			 * and this may be subject to breakage if those integer values
			 * change */
			inst = stappend(inst, "Search for this word or part:");
			inst = stappend(inst, "Search for this phrase:");
			inst = stappend(inst, "Search for multiple words:");
			inst = stappend(inst, "Search for attribute (eg Strongs#):");
			inst = stappend(inst, "SPACE");
			inst = stappend(inst, "Restrict search to this text:");

			const char* title = "Search";
			const char* secondinst = "Enter your search on the correct line"
				" and specify the scope if needed.";

			starray ret = showForm(title, inst, secondinst, floaty,
														 floatx, floatheight, floatwidth);
			if (ret.length == -1) {
				doresize(p);
				break;
			}
			if (ret.length != 0) {
				/* user entered something - get first field with data and
				 * the search scope (the last item), if any */
				int i = 0;
				for (i = 0; i < ret.length-1; i++) {
					// user entered a scope but no search
					//if (i == ret.length - 1) break;

					if (strcmp((ret.strings)[i], "") != 0) {
						focustab->setKey((ret.strings)[i]);
						focustab->setSearch(i * -1,
																(ret.strings)[ret.length-1]);
						break;
					}

				}

				if (i != ret.length - 1)
					focustab->loadText(scab.search(focustab->getModkey()));
			}
			wipeit(p);
			break; }

		case 't': {
			// toggle Strong's numbers
			int strongs = ! parseConf(config["markup"]["strongs"]);
			config["markup"]["strongs"] = strongs + '0';
			wipeit(p);
			saveit(configfile);
			break; }

		case 'w': {
			// toggle raw text
			int raw = ! parseConf(config["markup"]["rawtext"]);
			config["markup"]["rawtext"] = raw + '0';
			wipeit(p);
			saveit(configfile);
			break; }

		case '?': {
			// display help text
			clear();
			addstr(HELPTEXT);
			getch();
			wipeit(p);
			break; }

		case 'q':
			// quit - we shouldn't get here
			break;

		} // switch
	} // input loop

	// go away cleanly
	wrapup(0, NULL);

	// shouldn't get here
	return 0;
}
